unit uManipBMPV4;

{.$MODE Delphi}
{$mode objfpc}{$H+}

interface

uses LCLIntf, LCLType, LMessages, Graphics, Dialogs, SysUtils, Math;

function ColorVersClQuad(cl: tColor): tRGBQuad;
function clQuadVersColor(c3: TRGBQuad): tColor;
function clQuadEgales(c1, c2: tRGBQuad): boolean; // True si cl1 = cl2
function clQuadMix2(c1, c2: tRGBQuad): tRGBQuad; // Renvoie la moyenne des deux couleurs
function clQuadMix2K(const c1, c2: tRGBQuad; const kc1: Single): tRGBQuad;
//<-     renvoie la couleur dégradée égale à kc1 * c1 + (1 - kc1) * c2

procedure RGBtoHSL3(rgb: TRGBTriple; var H, S, L: double); // RGB dans 0..255, H,S,L dans 0..1
//        pour R = 184, G = 151, B =52, renvoie
//             H = 0,125 (et x 360 = 45) (ou x 240 = 30 Idem Delphi)
//             S = 0,559322033898305 (et x 240 = 134 Idem Delphi)
//             L = 0,462745098039216 (et x 240 = 111 Idem Delphi)
procedure HSLtoRGB3(H, S, L: double; var rgb: TRGBTriple);

procedure BmpDeltaLuminosite(const BmpS: tBitMap; kLum: single);
//        Retourne le Bmp dont la luminositié est multipliéepar kLum inférieur ou supérieur à 1.0

procedure BmpSatFoisK(const bmp: tBitMap; kSat: single);
//        Retourne le Bmp en modifiant la Saturation de ses couleurs de la valeur de kSat = par exemple à 1/1,1 ou 1,1

procedure DeltaRGB_Bmp(Bmp: tBitMap; kRouge, kVert, kBleu: single; PreserverGris: boolean);
//        Multiplie les composantes R,G,B par kRouge, kVert et kBleu
//        sauf dans les nuances de Gris si PreserverGris = True.

procedure BmpContraster(Bmp: tBitMap; CoefContraste: Single);
//        Retourne le Bmp dont le contraste est augmenté ou diminué : CoefContraste dans [-1,0..+1,0]

type
  TObjectProc = procedure of object;
procedure BmpFlouXY4(Bmp: tBitMap; TraineeX, TraineeY: byte; const ProgressCallBack: TObjectProc = nil);
//        Retourne le Bmp en lui ajoutant du Flou sur TraineeX, TraineeY pixels avec moyenne sur 4 pixels

implementation

function clQuadEgales(c1, c2: tRGBQuad): boolean;
begin Result := (c1.rgbRed = c2.rgbRed) and (c1.rgbGreen = c2.rgbGreen) and (c1.rgbBlue = c2.rgbBlue);
end;

function clQuadVersColor(c3: TRGBQuad): tColor;
begin Result := RGB(c3.rgbRed, c3.rgbGreen, c3.rgbBlue);
end;

function ColorVersClQuad(cl: tColor): tRGBQuad;
begin
  with Result do begin
    rgbRed := GetRValue(cl);
    rgbGreen := GetGValue(cl);
    rgbBlue := GetBValue(cl);
  end;
end;

function clQuadMix2(c1, c2: tRGBQuad): tRGBQuad;
begin
  with Result do begin
    rgbRed := (c1.rgbRed + c2.rgbRed) shr 1;
    rgbGreen := (c1.rgbGreen + c2.rgbGreen) shr 1;
    rgbBlue := (c1.rgbBlue + c2.rgbBlue) shr 1;
  end;
end;

function clQuadMix2K(const c1, c2: tRGBQuad; const kc1: Single): tRGBQuad;
begin
  with Result do begin
    rgbRed := round(c1.rgbRed * kc1 + c2.rgbRed * (1.0 - kc1));
    rgbGreen := round(c1.rgbGreen * kc1 + c2.rgbGreen * (1.0 - kc1));
    rgbBlue := round(c1.rgbBlue * kc1 + c2.rgbBlue * (1.0 - kc1));
  end;
end;

function clQuadMix4(c1, c2, c3, c4: tRGBQuad): tRGBQuad; // moyenne des 4 couleurs
begin
  with Result do begin
    rgbRed := (c1.rgbRed + c2.rgbRed + c3.rgbRed + c4.rgbRed) shr 2;
    rgbGreen := (c1.rgbGreen + c2.rgbGreen + c3.rgbGreen + c4.rgbGreen) shr 2;
    rgbBlue := (c1.rgbBlue + c2.rgbBlue + c3.rgbBlue + c4.rgbBlue) shr 2;
  end;
end;

// HSL <-> RGB ---------------------------------------------------------------------

procedure RGBtoHSL3(rgb: TRGBTriple; var H, S, L: double);
var delta, r, g, b, cmax, cmin: double;
begin
  r := rgb.rgbtRed / 255;
  g := rgb.rgbtGreen / 255;
  b := rgb.rgbtBlue / 255;

  if (r > b) and (r > g) then cmax := r
  else if g > b then cmax := g
  else cmax := b;

  if (r < b) and (r < g) then cmin := r
  else if g < b then cmin := g
  else cmin := b;

  L := (cmax + cmin) / 2.0;

  if cmax = cmin then begin
    S := 0;
    H := 0; // Teinte indéfinie
  end else begin
    delta := cmax - cmin;
    if L <= 0.5 then s := delta / (cmax + cmin)
    else s := delta / (2.0 - cmax - cmin);

    if r = cmax then H := (g - b) / delta
    else if g = cmax then H := 2.0 + (b - r) / delta
    else H := 4.0 + (r - g) / delta;
    H := H / 6.0;
    if H < 0 then H := H + 1;
  end;
end; // RGBtoHSL3

procedure HSLtoRGB3(H, S, L: double; var rgb: TRGBTriple);
var r, g, b, m1, m2: double;

  function HuetoRGB(m1, m2, h: double): double;
  begin if h < 0 then h := h + 1.0
    else if h > 1 then h := h - 1.0;

    if 6.0 * h < 1 then result := (m1 + (m2 - m1) * h * 6.0)
    else if 2.0 * h < 1 then result := m2
    else if 3.0 * h < 2.0 then result := (m1 + (m2 - m1) * ((2.0 / 3.0) - h) * 6.0)
    else result := m1;
  end;

begin if S = 0 then begin
    r := L;
    g := L;
    b := L;
  end else begin
    if L <= 0.5 then m2 := L * (1.0 + S)
    else m2 := L + S - L * S;
    m1 := 2.0 * L - m2;
    r := HuetoRGB(m1, m2, H + 1.0 / 3.0);
    g := HuetoRGB(m1, m2, H);
    b := HuetoRGB(m1, m2, H - 1.0 / 3.0);
  end;
  rgb.rgbtBlue := round(b * 255);
  rgb.rgbtGreen := round(g * 255);
  rgb.rgbtRed := round(r * 255);
end; //HSLtoRGB3

type pLigScan = ^tLigScan;
  tLigScan = array[Word] of TRGBTriple;

procedure BmpSatFoisK(const bmp: tBitMap; kSat: single);
var Lig: pLigScan;
  x, y: integer; H, S, L: double;
begin bmp.PixelFormat := pf24bit;
  for y := 0 to bmp.height - 1 do begin
    Lig := bmp.scanline[y];
    for x := 0 to bmp.width - 1 do begin
      RGBtoHSL3(tRGBTriple(Lig^[x]), H, S, L);
      if S * kSat <= 1.0 then S := S * kSat;
      HSLtoRGB3(H, S, L, tRGBTriple(Lig^[x]));
    end;
  end;
end; // BmpSatFoisK

procedure BmpContraster(Bmp: tBitMap; CoefContraste: Single);
type
  TRGBArray = array[0..10000] of TRGBTriple;
  pTRGBArray = ^TRGBArray;
  TArrayLigne = array[0..10000] of pTRGBArray;
var
  x, y: integer;
  R, G, B, pivotR, pivotG, pivotB: integer;
  Ligne: pTRGBArray;

  procedure PivotsContraste(var pivotR, pivotG, pivotB: integer);
  var x, y: integer; Rmax, Gmax, Bmax, Rmin, Gmin, Bmin: byte;
  begin
    Rmax := 0; Gmax := 0; Bmax := 0; Rmin := 255; Gmin := 255; Bmin := 255;
    for y := 0 to Bmp.Height - 1 do begin
      Ligne := Bmp.ScanLine[y];
      for x := 0 to Bmp.Width - 1 do begin
        R := Ligne^[x].rgbtRed;
        G := Ligne^[x].rgbtGreen;
        B := Ligne^[x].rgbtBlue;
        Rmax := max(R, Rmax); Gmax := max(G, Gmax); Bmax := max(B, Bmax);
        Rmin := min(R, Rmin); Gmin := min(G, Gmin); Bmin := min(B, Bmin);
      end;
    end;
    pivotR := (Rmax + Rmin) shr 1; pivotG := (Gmax + Gmin) shr 1; pivotB := (Bmax + Bmin) shr 1;
  end;

begin
  Bmp.PixelFormat := pf24bit;
  PivotsContraste(pivotR, pivotG, pivotB);
  for y := 0 to Bmp.Height - 1 do
  begin
    Ligne := Bmp.ScanLine[y];
    for x := 0 to Bmp.Width - 1 do
    begin
      R := Ligne^[x].rgbtRed;
      G := Ligne^[x].rgbtGreen;
      B := Ligne^[x].rgbtBlue;

      R := round(R + ((R - pivotR) * CoefContraste));
      G := round(G + ((G - pivotG) * CoefContraste));
      B := round(B + ((B - pivotB) * CoefContraste));

      if R < 0 then R := 0 else if R > 255 then R := 255;
      if G < 0 then G := 0 else if G > 255 then G := 255;
      if B < 0 then B := 0 else if B > 255 then B := 255;

      Ligne^[x].rgbtRed   := R;
      Ligne^[x].rgbtGreen := G;
      Ligne^[x].rgbtBlue  := B;
    end;
  end;
end;

procedure BmpFlouXY4(Bmp: tBitMap; TraineeX, TraineeY: byte; const ProgressCallBack: TObjectProc = nil);
// Flou avec moyenne sur 4 pixels
type tScanYX = array of array of integer; // Tableaux des adresses des pixels
var X, Y, Z, WS, HS: Integer; pixS: tScanYX;
  cN, cS, cE, cO, clMix: tRGBQuad;

  procedure Initialisations; // Initialisation du BitMap
  var x, y: integer;
    Scan0: Integer; // Valeur du pointeur d'entrée dans le Bitmap.
    MLS: Integer; //   Memory Line Size (en bytes) du Bitmap.
    Bpp: Integer; //   Bytes par pixel des Bitmaps.
  begin
    // BitMap-Source
    Bmp.PixelFormat := pf32bit; Bpp := 4; Scan0 := Integer(Bmp.ScanLine[0]); MLS := Integer(Bmp.ScanLine[1]) - Scan0;
    WS := Bmp.Width; HS := Bmp.Height;
    SetLength(pixS, HS, WS);
    for y := 0 to HS - 1 do begin
      for x := 0 to WS - 1 do pixS[y, x] := Scan0 + y * MLS + x * Bpp;
    end;
  end;

begin
  if (TraineeX = 0) and (TraineeY = 0) then EXIT;

  Initialisations;

  for Z := 1 to TraineeX do // décalage de trainéeX pixels :
  begin for Y := 0 to HS - 1 do
    begin
      for X := 0 to WS - TraineeX - 1 do begin
        cN := PRGBQuad(pixS[Y, X])^; cS := cN; cE := cN; cO := cN;
        if y > 0 then cN := PRGBQuad(pixS[Y - 1, X])^;
        if y < HS - 1 then cS := PRGBQuad(pixS[Y + 1, X])^;
        if x > 0 then cE := PRGBQuad(pixS[Y, X - 1])^;
        if x < WS - 1 then cO := PRGBQuad(pixS[Y, X + 1])^;
        clMix := clQuadMix4(cN, cS, cE, cO); // moyenne des 4 pixels
        PRGBQuad(pixS[y, x])^ := clMix;
      end;
      if Assigned(ProgressCallBack) then ProgressCallBack;
    end;
  end;
  if TraineeY = 0 then begin SetLength(pixS, 0, 0); EXIT; end;
  for Z := 0 to TraineeY - 1 do // décalage de trainéeY pixels :
  begin
    for y := 0 to HS - TraineeY - 1 do
    begin
      for X := 0 to WS - 1 do begin
        cN := PRGBQuad(pixS[Y, X])^; cS := cN; cE := cN; cO := cN;
        if y > 0 then cN := PRGBQuad(pixS[Y - 1, X])^;
        if y < HS - 1 then cS := PRGBQuad(pixS[Y + 1, X])^;
        if x > 0 then cE := PRGBQuad(pixS[Y, X - 1])^;
        if x < WS - 1 then cO := PRGBQuad(pixS[Y, X + 1])^;
        clMix := clQuadMix4(cN, cS, cE, cO); // moyenne des 4 pixels
        PRGBQuad(pixS[y, x])^ := clMix;
      end;
      if Assigned(ProgressCallBack) then ProgressCallBack;
    end;
  end;
  SetLength(pixS, 0, 0);
end; // BmpFlouXY4

procedure DeltaRGB_Bmp(Bmp: tBitMap; kRouge, kVert, kBleu: single; PreserverGris: boolean);
//        Multiplie les composantes R,G,B par kRouge, kVert et kBleu
//        sauf dans les nuances de Gris si PreserverGris = True.
type tScanYX = array of array of integer; // Tableaux des adresses des pixels
var X, Y, WS, HS, R, G, B: Integer; pixS: tScanYX;

  procedure Initialisations; // Initialisation du BitMap
  var x, y: integer;
    Scan0: Integer; // Valeur du pointeur d'entrée dans le Bitmap.
    MLS: Integer; //   Memory Line Size (en bytes) du Bitmap.
    Bpp: Integer; //   Bytes par pixel des Bitmaps.
  begin
    // BitMap-Source
    Bmp.PixelFormat := pf32bit; Bpp := 4; Scan0 := Integer(Bmp.ScanLine[0]); MLS := Integer(Bmp.ScanLine[1]) - Scan0;
    WS := Bmp.Width; HS := Bmp.Height;
    SetLength(pixS, HS, WS);
    for y := 0 to HS - 1 do begin
      for x := 0 to WS - 1 do pixS[y, x] := Scan0 + y * MLS + x * Bpp;
    end;
  end;

begin
  kRouge := abs(kRouge); kVert := abs(kVert); kBleu := abs(kBleu);
  if (kRouge = 1.0) and (kVert = 1.0) and (kBleu = 1.0) then EXIT;

  Initialisations;

  for Y := 0 to HS - 1 do
  begin
    for X := 0 to WS - 1 do begin
      with PRGBQuad(pixS[Y, X])^ do begin
        if (PreserverGris) and (rgbRed = rgbGreen) and (rgbGreen = rgbBlue) then Continue
        else begin;
          R := round(kRouge * rgbRed); G := round(kVert * rgbGreen); B := round(kBleu * rgbBlue);
          if R > 255 then R := 255; if G > 255 then G := 255; if B > 255 then B := 255;
          rgbRed := R; rgbGreen := G; rgbBlue := B;
        end;
      end;
    end;
  end;
  SetLength(pixS, 0, 0);
end; // DeltaRGB_Bmp

procedure BmpDeltaLuminosite(const BmpS: tBitMap; kLum: single);
//        Retourne le Bmp dont la luminositié est multipliée par kLum
type tScanYX = array of array of integer; // Tableaux des adresses des pixels
var X, Y, WS, HS, R, G, B: Integer; pixS: tScanYX;

  procedure Initialisations; // Initialisation du BitMap
  var x, y: integer;
    Scan0: Integer; // Valeur du pointeur d'entrée dans le Bitmap.
    MLS: Integer; //   Memory Line Size (en bytes) du Bitmap.
    Bpp: Integer; //   Bytes par pixel des Bitmaps.
  begin
    // BitMap-Source
    BmpS.PixelFormat := pf32bit; Bpp := 4; Scan0 := Integer(BmpS.ScanLine[0]); MLS := Integer(BmpS.ScanLine[1]) - Scan0;
    WS := BmpS.Width; HS := BmpS.Height;
    SetLength(pixS, HS, WS);
    for y := 0 to HS - 1 do begin
      for x := 0 to WS - 1 do pixS[y, x] := Scan0 + y * MLS + x * Bpp;
    end;
  end;

begin
  kLum := abs(kLum);
  if (kLum = 1.0) then EXIT;

  Initialisations;

  for Y := 0 to HS - 1 do
  begin
    for X := 0 to WS - 1 do begin
      with PRGBQuad(pixS[Y, X])^ do begin
        R := round(kLum * rgbRed); G := round(kLum * rgbGreen); B := round(kLum * rgbBlue);
        if R > 255 then R := 255; if G > 255 then G := 255; if B > 255 then B := 255;
        rgbRed := R; rgbGreen := G; rgbBlue := B;
      end;
    end;
  end;
  SetLength(pixS, 0, 0);
end; // BmpDeltaLuminosite

end.

